from __future__ import annotations

from datetime import timedelta

import httpx

from .. import models
from ..mockish import Mock


class Response(httpx.Response, models.Response):
    def __init__(
        self,
        status_code: int | None = None,
        headers: dict[str, str] | None = None,
        content: str | None = None,
        content_type: str | None = None,
        encoding: str | None = None,
        elapsed: timedelta | None = None,
        _data: models.ResponseData | None = None,
    ):
        """

        Args:
            status_code:
            headers:
            content:
            content_type:
            encoding:
            elapsed:
            _data:

        """
        super().__init__(status_code=200)

        if not _data:
            _data = super()._prepare_response_data(
                status_code=status_code,
                headers=headers,
                content=content,
                content_type=content_type,
                encoding=encoding,
                elapsed=elapsed,
            )

        self.status_code = _data.status_code

        self.headers = httpx.Headers(_data.headers)

        if _data.content:
            self._content = _data.content

        if _data.elapsed:
            self.elapsed = _data.elapsed

        self._request = Mock(spec_set=httpx.Request)



from __future__ import annotations

import json
from random import randint
from typing import Any
from unittest import mock
from uuid import uuid4

import pytest
import requests

import mockish
from mockish.constants import CONTENT_TYPE_DEFAULT, CONTENT_TYPE_JSON


@pytest.mark.parametrize(
    ("content", "content_type"),
    [
        ("asdf", None),
        ("asdf", "asdf"),
        (json.dumps({"hello": "asdf"}), None),
        (json.dumps({"hello": "asdf"}), CONTENT_TYPE_JSON),
    ],
)
def test_mock_response(content: str, content_type: str | None) -> None:
    expected_content: str = content
    expected_content_type: str | None = content_type
    expected_status_code: int = 200 + randint(0, 99)
    expected_headers: dict[str, str] = {"User-Agent": str(uuid4())}

    mock_resp: requests.Response = mockish.requests.Response(
        content=expected_content,
        content_type=expected_content_type,
        status_code=expected_status_code,
        headers=expected_headers,
    )

    assert isinstance(mock_resp, requests.Response)

    assert "Content-Type" in mock_resp.headers
    assert "Content-Length" in mock_resp.headers

    assert mock_resp.text == expected_content
    assert mock_resp.content == expected_content.encode()
    assert int(mock_resp.headers["Content-Length"]) == len(mock_resp.content)

    if not expected_content_type:
        expected_content_type = CONTENT_TYPE_DEFAULT
    elif expected_content_type == CONTENT_TYPE_JSON:
        assert mock_resp.json() == json.loads(expected_content)

    assert mock_resp.headers["Content-Type"] == expected_content_type

    assert mock_resp.encoding is None

    assert expected_headers
    for k, v in expected_headers.items():
        assert mock_resp.headers[k] == v
        assert mock_resp.headers[k.lower()] == v
        assert mock_resp.headers[k.upper()] == v


def test_mock_response_from_dict() -> None:
    mock_resp: requests.Response = mockish.requests.Response.from_dict(
        {"hello": str(uuid4())},
    )

    assert isinstance(mock_resp, requests.Response)

    assert "Content-Length" in mock_resp.headers
    assert "Content-Type" in mock_resp.headers

    test_mock_response(
        content=mock_resp.text,
        content_type=mock_resp.headers["Content-Type"],
    )


def _assert_mock_response_values(
    mock_session,
    responses: list[requests.Response],
) -> None:
    for expected in responses:
        fake_endpoint: str = str(uuid4())
        observed: requests.Response = mock_session.get(fake_endpoint)

        mock_session.get.assert_called_with(fake_endpoint)

        assert isinstance(observed, requests.Response)
        assert observed == expected

    assert mock_session.get.call_count == len(responses)


def test_mock_request_method_return_value() -> None:
    n_calls: int = 5

    mock_resp: requests.Response = mockish.requests.Response.from_dict(
        content={"hello": str(uuid4())},
        status_code=200 + randint(0, 99),
    )

    mock_session: mock.Mock = mockish.Mock(
        spec_set=requests.Session,
        **{"get": mockish.Mock(return_value=mock_resp)},
    )

    _assert_mock_response_values(
        mock_session=mock_session,
        responses=[mock_resp] * n_calls,
    )

    try:
        mock_session.get(...)
    except StopIteration:
        pytest.fail("this method should return the value indefinitely")


def test_mock_request_method_return_values() -> None:
    n_responses: int = 5

    mock_resp_list: list[requests.Response] = [
        mockish.requests.Response.from_dict(
            content={"hello": str(uuid4())},
            status_code=200 + randint(0, 99),
        )
        for _ in range(n_responses)
    ]

    mock_session: mock.Mock = mockish.Mock(
        spec_set=requests.Session,
        **{"get": mockish.Mock(return_each=mock_resp_list)},
    )

    _assert_mock_response_values(mock_session=mock_session, responses=mock_resp_list)

    with pytest.raises(StopIteration):
        mock_session.get(...)


def test_mock_request_method_callable() -> None:
    n_calls = 5

    mock_resp: requests.Response = mockish.requests.Response.from_dict(
        {"hello": str(uuid4())},
    )

    def my_func(
        *args: Any,  # pylint: disable=unused-argument
        **kwargs: Any,  # pylint: disable=unused-argument
    ) -> requests.Response:
        return mock_resp

    mock_session: mock.Mock = mockish.Mock(
        spec_set=requests.Session,
        **{"get": mockish.Mock(return_call=my_func)},
    )

    _assert_mock_response_values(
        mock_session=mock_session,
        responses=[mock_resp] * n_calls,
    )

    try:
        mock_session.get(...)
    except StopIteration:
        pytest.fail("method should return the value indefinitely")



from __future__ import annotations

import json
from random import randint
from typing import Any
from uuid import uuid4

import httpx
import pytest

import mockish
from mockish.constants import CONTENT_TYPE_DEFAULT, CONTENT_TYPE_JSON


@pytest.mark.parametrize(
    ("content", "content_type"),
    [
        ("asdf", None),
        ("asdf", "asdf"),
        (json.dumps({"hello": "asdf"}), None),
        (json.dumps({"hello": "asdf"}), CONTENT_TYPE_JSON),
    ],
)
def test_mock_response(content: str, content_type: str | None) -> None:
    expected_content: str = content
    expected_content_type: str | None = content_type
    expected_status_code: int = 200 + randint(0, 99)
    expected_headers: dict[str, str] = {"User-Agent": str(uuid4())}

    mock_resp: httpx.Response = mockish.httpx.Response(
        content=expected_content,
        content_type=expected_content_type,
        status_code=expected_status_code,
        headers=expected_headers,
    )

    assert isinstance(mock_resp, httpx.Response)

    assert "Content-Type" in mock_resp.headers
    assert "Content-Length" in mock_resp.headers

    assert mock_resp.text == expected_content
    assert mock_resp.content == expected_content.encode()
    assert int(mock_resp.headers["Content-Length"]) == len(mock_resp.content)

    if not expected_content_type:
        expected_content_type = CONTENT_TYPE_DEFAULT
    elif expected_content_type == CONTENT_TYPE_JSON:
        assert mock_resp.json() == json.loads(expected_content)

    assert mock_resp.headers["Content-Type"] == expected_content_type

    assert mock_resp.encoding == "utf-8"

    assert expected_headers
    for k, v in expected_headers.items():
        assert mock_resp.headers[k] == v
        assert mock_resp.headers[k.lower()] == v
        assert mock_resp.headers[k.upper()] == v


def test_mock_response_from_dict() -> None:
    mock_resp: httpx.Response = mockish.httpx.Response.from_dict(
        {"hello": str(uuid4())},
    )

    assert isinstance(mock_resp, httpx.Response)

    assert "Content-Length" in mock_resp.headers
    assert "Content-Type" in mock_resp.headers

    test_mock_response(
        content=mock_resp.text,
        content_type=mock_resp.headers["Content-Type"],
    )


def _assert_mock_response_values(
    mock_session,
    responses: list[httpx.Response],
) -> None:
    for expected in responses:
        fake_endpoint: str = str(uuid4())
        observed: httpx.Response = mock_session.get(fake_endpoint)

        mock_session.get.assert_called_with(fake_endpoint)

        assert isinstance(observed, httpx.Response)
        assert observed == expected

    assert mock_session.get.call_count == len(responses)


def test_mock_request_method_return_value() -> None:
    n_calls: int = 5

    mock_resp: httpx.Response = mockish.httpx.Response.from_dict(
        content={"hello": str(uuid4())},
        status_code=200 + randint(0, 99),
    )

    with mockish.patch.object(httpx, "get", mockish.Mock(return_value=mock_resp)):
        _assert_mock_response_values(
            mock_session=httpx,
            responses=[mock_resp] * n_calls,
        )

        try:
            httpx.get(...)
        except StopIteration:
            pytest.fail("this method should return the value indefinitely")


def test_mock_request_method_return_values() -> None:
    n_responses: int = 5

    mock_resp_list: list[httpx.Response] = [
        mockish.httpx.Response.from_dict(
            content={"hello": str(uuid4())},
            status_code=200 + randint(0, 99),
        )
        for _ in range(n_responses)
    ]

    with mockish.patch.object(httpx, "get", mockish.Mock(return_each=mock_resp_list)):
        _assert_mock_response_values(mock_session=httpx, responses=mock_resp_list)

        with pytest.raises(StopIteration):
            httpx.get(...)


def test_mock_request_method_callable() -> None:
    n_calls = 5

    mock_resp: httpx.Response = mockish.httpx.Response.from_dict(
        {"hello": str(uuid4())},
    )

    def my_func(
        *args: Any,  # pylint: disable=unused-argument
        **kwargs: Any,  # pylint: disable=unused-argument
    ) -> httpx.Response:
        return mock_resp

    with mockish.patch.object(httpx, "get", mockish.Mock(return_call=my_func)):
        _assert_mock_response_values(
            mock_session=httpx,
            responses=[mock_resp] * n_calls,
        )

        try:
            httpx.get(...)
        except StopIteration:
            pytest.fail("method should return the value indefinitely")



from typing import Any, Callable, Dict, Optional, Sequence
from unittest import mock

from pydantic import validate_arguments

__all__ = [
    "Mock",
    "AsyncMock",
    "patch_fastapi_dependencies",
]


class _AsyncMock(mock.Mock):
    # AsyncMock for Python 3.7
    # (added to stdlib in Python 3.8)

    # pylint: disable=invalid-overridden-method, useless-parent-delegation
    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return super().__call__(*args, **kwargs)


@validate_arguments(config={"arbitrary_types_allowed": True})
def _build_mock(
    *,
    is_async: bool,
    return_value: Optional[Any] = None,
    return_call: Optional[Callable[..., Optional[Any]]] = None,
    return_once: Optional[Any] = None,
    return_each: Optional[Sequence[Any]] = None,
    return_exception: Optional[Exception] = None,
    side_effect: Optional[Any] = None,
    **kwargs: Any,
) -> mock.Mock:
    if (
        sum(
            [
                bool(return_value),
                bool(return_once),
                bool(return_each),
                bool(return_call),
                bool(return_exception),
                bool(side_effect),
            ],
        )
        > 1
    ):
        raise ValueError("Specify exactly one argument.")

    if side_effect:
        kwargs["side_effect"] = side_effect

    elif return_value:
        kwargs["side_effect"] = lambda *args, **kwargs: return_value

    elif return_call:
        kwargs["side_effect"] = return_call

    elif return_once:
        kwargs["side_effect"] = [return_once]

    elif return_each:
        kwargs["side_effect"] = return_each

    elif return_exception:
        kwargs["side_effect"] = return_exception

    return _AsyncMock(**kwargs) if is_async else mock.Mock(**kwargs)


def Mock(
    *,
    return_value: Optional[Any] = None,
    return_call: Optional[Callable[..., Optional[Any]]] = None,
    return_once: Optional[Any] = None,
    return_each: Optional[Sequence[Any]] = None,
    return_exception: Optional[Exception] = None,
    **kwargs: Any,
) -> mock.Mock:
    """A thin wrapper around [unittest.mock.Mock](https://docs.python.org/3/library/unittest.mock.html#the-mock-class) to abstract away the use of `side_effect` in favor of these explicit `return_X` parameters:

    Args:
        return_value: return the given value
        return_call: return the value returned by the given callable
        return_once: return the given value exactly once
        return_each: consecutively return each element of the given iterable
        return_exception: raise the given exception

    Returns:
        : A `Mock` object

    Examples:
        - `return_value`
        >>> obj = Mock(return_value='hello world')
        >>> obj()
        'hello world'

        - `return_call`
        >>> func = lambda: 'hello world'
        >>> obj = Mock(return_call=func)
        >>> obj()
        'hello world'

        - `return_once`
        >>> obj = Mock(return_once='hello world')
        >>> obj()
        'hello world'
        >>> obj()
        Traceback (most recent call last):
        ...
        StopIteration

        - `return_each`
        >>> obj = Mock(return_each=[1, 2, 3])
        >>> obj()
        1
        >>> obj()
        2
        >>> obj()
        3
        >>> obj()
        Traceback (most recent call last):
        ...
        StopIteration

        - `return_exception`:
        >>> obj = Mock(return_exception=ValueError("hello world"))
        >>> obj()
        Traceback (most recent call last):
        ...
        ValueError: hello world
    """
    return _build_mock(
        is_async=False,
        return_value=return_value,
        return_call=return_call,
        return_once=return_once,
        return_each=return_each,
        return_exception=return_exception,
        **kwargs,
    )


def AsyncMock(
    *,
    return_value: Optional[Any] = None,
    return_call: Optional[Callable[..., Optional[Any]]] = None,
    return_once: Optional[Any] = None,
    return_each: Optional[Sequence[Any]] = None,
    return_exception: Optional[Exception] = None,
    **kwargs: Any,
) -> mock.Mock:
    """Same as `mockish.Mock`, but returns an async `Mock`.

    Returns:
        : An async `Mock` object
    """
    return _build_mock(
        is_async=True,
        return_value=return_value,
        return_call=return_call,
        return_once=return_once,
        return_each=return_each,
        return_exception=return_exception,
        **kwargs,
    )


def patch_fastapi_dependencies(
    *args: Any,
    overrides: Optional[Dict[Callable[..., Any], Callable[..., Any]]],
    remove: bool = False,
) -> None:
    """Recursively patch dependencies of FastAPI instance(s).

    [Read about FastAPI test dependencies.](https://fastapi.tiangolo.com/advanced/testing-dependencies/)
    > Note: `fastapi` must be installed.

    Args:
        *args: One or more FastAPI instances
        overrides: A mapping of overrides, or `None` to clear all
        remove: Remove the provided overrides

    Raises:
        TypeError: raised if any of `args` is not a FastAPI instance

    Returns:
        : No value returned

    Examples:
        ```py
        from fastapi import FastAPI
        from fastapi.testclient import TestClient

        app = FastAPI(...)

        mockish.patch_fastapi_dependencies(
            app,
            overrides={
                get_app_settings: lambda: app_settings,
            },
        )

        @app.get("/")
        def get() -> dict[str, str]:
            return {"hello": "world"}

        return TestClient(app)
        ```
    """
    from fastapi import FastAPI  # pylint: disable=import-error
    from starlette.routing import Mount  # pylint: disable=import-error

    for x in args:
        if not isinstance(x, FastAPI):
            raise TypeError(f"Expected type 'FastAPI'; given: {type(x)}")

        if overrides is None:
            x.dependency_overrides.clear()
        elif remove:
            for k in overrides:
                del x.dependency_overrides[k]
        else:
            x.dependency_overrides.update(overrides)

        patch_fastapi_dependencies(
            *[
                y.app
                for y in x.routes
                if isinstance(y, Mount) and isinstance(y.app, FastAPI)
            ],
            overrides=overrides,
            remove=remove,
        )



from __future__ import annotations

from unittest.mock import patch

from . import httpx, requests
from .__version__ import __version__
from .mockish import AsyncMock, Mock, patch_fastapi_dependencies

__all__ = [
    "__version__",
    "Mock",
    "AsyncMock",
    "patch",
    "patch_fastapi_dependencies",
    "requests",
    "httpx",
]



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import timedelta
from typing import Any, TypeVar

from . import constants

T = TypeVar("T", bound="Response")


@dataclass
class ResponseData:
    status_code: int
    headers: dict[str, str]
    content: bytes | None
    elapsed: timedelta | None


class Response(ABC):
    @abstractmethod
    def __init__(
        self,
        status_code: int | None = None,
        headers: dict[str, str] | None = None,
        content: str | None = None,
        content_type: str | None = None,
        encoding: str | None = None,
        elapsed: timedelta | None = None,
        _data: ResponseData | None = None,
    ) -> None:
        ...

    @staticmethod
    def _prepare_response_data(
        status_code: int | None = None,
        headers: dict[str, str] | None = None,
        content: str | None = None,
        content_type: str | None = None,
        encoding: str | None = None,
        elapsed: timedelta | None = None,
    ) -> ResponseData:
        if not headers:
            headers = {}

        if content:
            if not content_type:
                content_type = constants.CONTENT_TYPE_DEFAULT

            content_encoded: bytes

            if encoding:
                content_encoded = content.encode(encoding)
                content_type += f"; charset={encoding}"
            else:
                content_encoded = content.encode()

            headers["Content-Type"] = content_type

            if "Content-Length" not in headers:
                headers["Content-Length"] = str(len(content_encoded))

        return ResponseData(
            status_code=int(status_code) if status_code else 200,
            headers=headers,
            content=content_encoded,
            elapsed=elapsed,
        )

    @classmethod
    def create(cls: type[T], data: ResponseData) -> T:
        return cls(_data=data)

    @classmethod
    def from_dict(cls: type[T], content: dict[str, Any], **kwargs: Any) -> T:
        return cls.create(
            cls._prepare_response_data(
                content=json.dumps(content, ensure_ascii=False),
                content_type="application/json",
                **kwargs,
            ),
        )

    @classmethod
    def from_file(cls: type[T], path: str, encoding: str = "utf-8", **kwargs: Any) -> T:
        with open(path, encoding=encoding) as f:
            return cls.create(
                cls._prepare_response_data(
                    content=f.read(),
                    content_type=(
                        "application/json" if path.lower().endswith(".json") else None
                    ),
                    **kwargs,
                ),
            )
